import logging
from typing import Any, Dict, Generic, Hashable, List, NewType, Optional, TypeVar, cast

from pydantic import BaseModel

T = TypeVar("T")

log = logging.getLogger(__name__)

Priority = NewType("Priority", int)


LOWER_PRIORITY = Priority(30)
LOW_PRIORITY = Priority(40)
NORMAL_PRIORITY = Priority(50)
HIGH_PRIORITY = Priority(60)
HIGHER_PRIORITY = Priority(70)


class MergeStrategy(Generic[T]):
    def __merge__(self, value1: T, value2: T, priority1: Priority, priority2: Priority) -> T:
        raise NotImplementedError


class SmartPreferTrustedStrategy(MergeStrategy[T]):
    def __merge__(
        self, target_value: T, source_value: T, target_priority: Priority, source_priority: Priority
    ) -> T:
        # If either value is None, return the non-None value
        if target_value is None:
            return source_value
        if source_value is None:
            return target_value

        if source_priority > target_priority:
            preferred, other = source_value, target_value
            high_priority, low_priority = source_priority, target_priority
        else:
            preferred, other = target_value, source_value
            high_priority, low_priority = target_priority, source_priority

        if hasattr(preferred, "__merge__") and callable(getattr(preferred, "__merge__")):
            return preferred.__merge__(other, high_priority, low_priority)  # type: ignore
        elif isinstance(preferred, list) and isinstance(other, list):
            return SmartListAsSetMergeStrategy().__merge__(
                preferred, other, high_priority, low_priority
            )
        elif isinstance(preferred, dict) and isinstance(other, dict):
            return SmartCombineDictsStrategy().__merge__(
                preferred, other, high_priority, low_priority
            )
        else:
            return preferred


class SmartListMergeStrategy(MergeStrategy[List[Any]]):
    def __merge__(
        self, value1: List[Any], value2: List[Any], priority1: Priority, priority2: Priority
    ) -> List[Any]:
        merged = []
        seen_complex: Dict[int, Any] = {}

        for item in value1 + value2:
            if isinstance(item, (int, str, float, bool)):
                merged.append(item)
            else:
                if hasattr(item, "equality_hash") and callable(getattr(item, "equality_hash")):
                    item_hash = item.equality_hash()
                    if item_hash in seen_complex:
                        existing = seen_complex[item_hash]
                        if hasattr(existing, "__merge__") and callable(
                            getattr(existing, "__merge__")
                        ):
                            merged[merged.index(existing)] = existing.__merge__(
                                item, priority1, priority2
                            )
                    else:
                        merged.append(item)
                        seen_complex[item_hash] = item
                elif hasattr(item, "__merge__") and callable(getattr(item, "__merge__")):
                    existing = next((x for x in merged if type(x) == type(item)), None)
                    if existing:
                        merged[merged.index(existing)] = existing.__merge__(
                            item, priority1, priority2
                        )
                    else:
                        merged.append(item)
                else:
                    merged.append(item)

        return merged


class SmartListAsSetMergeStrategy(MergeStrategy[List[Any]]):
    def __merge__(
        self,
        target_value: List[Any],
        source_value: List[Any],
        target_priority: Priority,
        source_priority: Priority,
    ) -> List[Any]:
        merged = []
        seen_complex = {}

        def add_item(item, priority):
            if isinstance(item, (int, str, float, bool)):
                if item not in merged:
                    merged.append(item)
            elif hasattr(item, "equality_hash") and callable(getattr(item, "equality_hash")):
                item_hash = item.equality_hash()
                if item_hash in seen_complex:
                    existing, existing_priority = seen_complex[item_hash]
                    if hasattr(existing, "__merge__") and callable(getattr(existing, "__merge__")):
                        merged_item = existing.__merge__(item, existing_priority, priority)
                        merged[merged.index(existing)] = merged_item
                        seen_complex[item_hash] = (merged_item, max(existing_priority, priority))
                else:
                    merged.append(item)
                    seen_complex[item_hash] = (item, priority)
            else:
                # For objects without equality_hash, we'll use their string representation as a hash
                item_str = str(item)
                if item_str not in seen_complex:
                    merged.append(item)
                    seen_complex[item_str] = (item, priority)

        for item in target_value:
            add_item(item, target_priority)

        for item in source_value:
            add_item(item, source_priority)

        return merged


class StringMergeStrategy(MergeStrategy[str]):
    def __merge__(self, value1: str, value2: str, priority1: Priority, priority2: Priority) -> str:
        print(f"Merging strings: '{value1}' and '{value2}'")  # Debug print
        if priority1 >= priority2:
            return value1
        else:
            return value2


class SmartCombineDictsStrategy(MergeStrategy[Dict[Any, Any]]):
    def __merge__(
        self,
        target_value: Dict[Any, Any],
        source_value: Dict[Any, Any],
        target_priority: Priority,
        source_priority: Priority,
    ) -> Dict[Any, Any]:
        log.debug(
            f"Merging dicts with priorities: target={target_priority}, source={source_priority}"
        )
        log.debug(f"Target dict: {target_value}")
        log.debug(f"Source dict: {source_value}")

        merged = {}
        all_keys = set(target_value.keys()) | set(source_value.keys())

        for key in all_keys:
            if key in target_value and key in source_value:
                log.debug(f"Merging key '{key}' present in both dicts")
                merged[key] = merge_value(
                    source_value=source_value[key],
                    target_value=target_value[key],
                    source_priority=source_priority,
                    target_priority=target_priority,
                )
            elif key in target_value:
                log.debug(f"Using target value for key '{key}'")
                merged[key] = target_value[key]
            else:
                log.debug(f"Using source value for key '{key}'")
                merged[key] = source_value[key]

        log.debug(f"Merged dict: {merged}")
        return merged


def get_default_strategy(field_type: type) -> MergeStrategy:
    if issubclass(field_type, list):
        return SmartListAsSetMergeStrategy()
    elif issubclass(field_type, dict):
        return SmartCombineDictsStrategy()
    else:
        return SmartPreferTrustedStrategy()


def merge_value(
    source_value: T,
    target_value: T,
    source_priority: Priority,
    target_priority: Priority,
    field_name: Optional[str] = None,
) -> T:
    log.debug(
        f"Merging values for field '{field_name}': target={target_value}, source={source_value}"
    )
    log.debug(f"Priorities: target={target_priority}, source={source_priority}")

    if target_value is None:
        log.debug("Target value is None, using source value")
        return source_value
    if source_value is None:
        log.debug("Source value is None, using target value")
        return target_value

    if isinstance(target_value, BaseModel) and isinstance(source_value, BaseModel):
        log.debug("Merging nested models")
        merged_model = target_value.model_copy()
        merge_model(
            source=source_value,
            target=merged_model,
            source_priority=source_priority,
            target_priority=target_priority,
        )
        return merged_model
    elif isinstance(target_value, dict) and isinstance(source_value, dict):
        log.debug("Merging dictionaries")
        return SmartCombineDictsStrategy().__merge__(
            target_value=target_value,
            source_value=source_value,
            target_priority=target_priority,
            source_priority=source_priority,
        )
    elif isinstance(target_value, list) and isinstance(source_value, list):
        log.debug("Merging lists")
        return SmartListAsSetMergeStrategy().__merge__(
            target_value=target_value,
            source_value=source_value,
            target_priority=target_priority,
            source_priority=source_priority,
        )
    else:
        if source_priority > target_priority:
            log.debug("Source priority higher, using source value")
            return source_value
        else:
            log.debug("Target priority higher or equal, using target value")
            return target_value


def merge_model(
    source: BaseModel, target: BaseModel, source_priority: Priority, target_priority: Priority
) -> None:
    for field_name in target.model_fields:
        if hasattr(source, field_name):
            target_value = getattr(target, field_name)
            source_value = getattr(source, field_name)

            if target_value is not None or source_value is not None:
                merged_value = merge_value(
                    target_value=target_value,
                    source_value=source_value,
                    target_priority=target_priority,
                    source_priority=source_priority,
                    field_name=field_name,
                )
                setattr(target, field_name, merged_value)



from typing import TYPE_CHECKING, Optional

from beanie import PydanticObjectId
from pydantic import Field

from qai.schema.models.addons import CreatedAtDoc, DateRange, Deleteable, Labels, Provenance, Taggable
from qai.schema.models.address_model import Address
from qai.schema.models.email_model import Email


class Job(CreatedAtDoc, Deleteable, Taggable, Labels, DateRange):
    title: Optional[str] = Field(default=None, description="The title of the job")
    division: Optional[str] = Field(default=None, description="The division of the job")
    company: Optional[str] = Field(default=None, description="The company of the job")
    company_id: Optional[PydanticObjectId] = Field(
        default=None, description="The reference to the company"
    )
    location: Optional[Address] = Field(default=None, description="The location of the job")
    description: Optional[str] = Field(default=None, description="The description of the job")
    notes: Optional[str] = Field(default=None, description="The notes of the job")
    current: bool = Field(default=False, description="Whether the job is the current one")
    email: Optional[Email] = Field(default=None, description="The email of the job")
    provenance: Optional[Provenance] = Field(default=None, description="The provenance")

    class Settings:
        equality_fields = ["title", "company_id"]
        keep_nulls = False


def _sort_job_key(obj: Job):
    ## sort work_history by start_date and then end_date
    end_value = obj.end if obj.end is not None else float("inf")
    return (obj.start, end_value)



from typing import TYPE_CHECKING, TypeAlias, TypeVar

from beanie import Document
from beanie import Link as BeanieLink
from beanie.odm.queries.find import FindOne as BeanieFindOne

if TYPE_CHECKING:
    _T = TypeVar("_T", bound=Document)
    Link: TypeAlias = _T
    FindOne: TypeAlias = _T
else:
    Link = BeanieLink
    FindOne = BeanieFindOne




from functools import partial
from typing import Any, Type, cast

from postal.parser import parse_address as postal_parse_address  # type: ignore

from qai.schema.extensions import ET
from qai.schema.models.address_model import Address as Address
import pycountry

def parse_address(s: str, *args, **kwargs) -> Address:
    """
    Transform an address string into a structured address object.
    Args:
        s: The address string to parse.
    Returns:
        An Address object.

    Fields inside parsed address:
    "house_number"
    "street"
    "street2"
    "city"
    "state"
    "postal_code"
    "country"

    """
    country_str = kwargs.get("country", None)
    if country_str:
        country = pycountry.countries.search_fuzzy(country_str)
        if country:
            kwargs["country"] = country[0].alpha_2 # type: ignore
    
    if "country" in kwargs:
        ## TODO, currently postal has some error where country="str" doesn't work
        ## So we need to pass the country as part of the string, if it's not in the string
        if not s.endswith(kwargs["country"]):
            s = f"{s}, {kwargs['country']}"
        kwargs.pop("country")

    parsed_address = postal_parse_address(s, **kwargs)

    address_dict: dict[str, Any] = {
        "current": False,
    }
    street_parts = []
    for component, label in parsed_address:
        if label in ("house_number", "road"):
            street_parts.append(component)
        elif label in ("city", "suburb"):
            address_dict["city"] = component
        elif label == "state":
            address_dict["state"] = component
        elif label == "postcode":
            address_dict["postal_code"] = component
        elif label == "country":
            address_dict["country"] = component
    if street_parts is not None:
        address_dict["street"] = " ".join(street_parts)
    ## Remove empy strings
    address_dict = {k: v for k, v in address_dict.items() if v}
    return Address(**address_dict)



from typing import Optional, List
from pydantic import BaseModel, Field
from pydantic import BaseModel
from typing import Any
from beanie import PydanticObjectId
from copy import deepcopy
from qai.schema import ExtendedDocument
from qai.schema import (
    Campaign,
    CampaignBatch,
    CampaignOptions,
    CampaignStep,
    Company,
    Contact,
    ExistingCustomer,
    OutreachType,
    ExcludeReason,
    Person,
    Provenance,
)


def resolve(p: Person, other: Person, source: Provenance) -> Person:
    if p.sources is None:
        p.sources = []
    for s in p.sources:
        if source.source == s.source:
            return p
        
    if not p.name and other.name:
        p.name = other.name
    if not p.gender and other.gender:
        p.gender = other.gender
    p.emails = merge_lists(p.emails, other.emails)
    # p.social_media = merge_lists(p.social_media, other.social_media)
    # p.phone_numbers = merge_lists(p.phone_numbers, other.phone_numbers)
    # p.addresses = merge_lists(p.addresses, other.addresses)
    p.work_history = merge_lists(p.work_history, other.work_history)
    if other.sources:
        for source in other.sources:
            if source not in p.sources:
                p.sources.append(source)
    p.sources.append(source)
    return p


def find_in_list(item: ExtendedDocument, item_list: list[ExtendedDocument]) -> bool:
    for i in item_list:
        if item.eq(i, to_lower=True, nones_ok=True):
            return True
    return False


def merge_lists(existing: Optional[list[Any]], new: Optional[list[Any]]) -> Optional[List[Any]]:
    if existing is None:
        existing = []
    if new:
        for item in new:
            if not find_in_list(item, existing):
                existing.append(item)
    return existing if existing else None



from enum import StrEnum


class ProvenanceType(StrEnum):
    OTHER = "other"
    MANUAL = "manual"
    AI_GENERATED = "ai_generated"


class ExcludeReason(StrEnum):
    ADDRESS_INVALID = "address_invalid"
    DOMAIN_INVALID = "domain_invalid"
    DOMAIN_FILTERED = "domain_filtered"
    EMAIL_BOUNCED = "email_bounced"
    EMAIL_FILTERED = "email_filtered"
    EMAIL_INVALID = "email_invalid"
    EMAIL_DOMAIN_FILTERED = "email_domain_filtered"
    EMAIL_UNVERIFIABLE = "email_unverifiable"
    NO_ADDRESS = "no_address"
    NO_COMPANY = "no_company"
    NO_COMPANY_DOMAIN = "no_company_domain"
    NO_EMAIL = "no_email"
    NO_JOB_TITLE = "no_job_title"
    NO_NAME = "no_name"
    NO_PERSON = "no_person"
    NO_PHONE = "no_phone"
    NO_SOCIAL_MEDIA = "no_social_media"
    PHONE_INVALID = "phone_invalid"
    SOCIAL_MEDIA_INVALID = "social_media_invalid"
    TITLE_FILTERED = "title_filtered"
    COMPANY_NAME_MISMATCH = "company_name_mismatch"

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


class OutreachType(StrEnum):
    OTHER = "other"
    EMAIL = "email"
    PHONE = "phone"
    SOCIAL_MEDIA = "social_media"
    MAIL = "mail"
    IN_PERSON = "in_person"
    WHATSAPP = "whatsapp"
    LINKEDIN = "linkedin"
    TWITTER = "twitter"
    FACEBOOK = "facebook"
    INSTAGRAM = "instagram"
    GITHUB = "github"
    ZOOM = "zoom"


class SexEnum(StrEnum):
    OTHER = "other"
    MALE = "male"
    FEMALE = "female"
    INTERSEX = "intersex"


class GenderEnum(StrEnum):
    OTHER = "other"
    MALE = "male"
    FEMALE = "female"
    NON_BINARY = "non-binary"
    GENDERQUEER = "genderqueer"
    AGENDER = "agender"
    GENDERFLUID = "genderfluid"



from typing import Optional

from nameparser import HumanName
from pydantic import BaseModel, Field

from qai.schema.models.addons import CreatedAtDoc, Provenance


class Name(BaseModel):
    title: Optional[str] = Field(
        default=None, description="The title of the person, Ex: Mr, Mrs, Dr, etc."
    )
    first: Optional[str] = Field(default=None, description="The first name")
    middle: Optional[str] = Field(default=None, description="The middle name")
    last: Optional[str] = Field(default=None, description="The last name")
    suffix: Optional[str] = Field(
        default=None, description="The suffix of the person, Ex: Jr, Sr, etc."
    )
    nickname: Optional[str] = Field(default=None, description="The nickname of the person")
    surnames: Optional[list[str]] = Field(default=None, description="The surnames of the person")
    provenance: Optional[Provenance] = Field(
        default=None, description="The provenance of the PhoneNumber"
    )

    @property
    def full_name(self) -> str:
        return str(
            HumanName(
                first=self.first,
                middle=self.middle,
                last=self.last,
                suffix=self.suffix,
                title=self.title,
            )
        )

    @staticmethod
    def from_str(name_str: str) -> "Name":
        if not name_str:
            raise ValueError(f"Name string '{name_str}' is empty.")

        name = HumanName(name_str)

        # Filter out empty strings from name attributes
        attributes = {
            "title": name.title,
            "first": name.first,
            "middle": name.middle,
            "last": name.last,
            "suffix": name.suffix,
            "nickname": name.nickname,
        }

        filtered_attributes = {key: value for key, value in attributes.items() if value}
        ## If everything is empty, return None
        if not filtered_attributes:
            raise ValueError(f"Name string '{name_str}' is empty.")
        surnames = [str(x) for x in name.surnames_list] if name.surnames_list else None

        return Name(
            **filtered_attributes,
            surnames=surnames,
            provenance=None
        )



from enum import StrEnum
from typing import Optional

from pydantic import EmailStr, Field

from qai.schema.extensions import ExtendedDocument

from qai.schema.models.addons import CreatedAtDoc, ExpiredAtDoc, Provenance


class EmailType(StrEnum):
    OTHER = "other"
    PERSONAL = "personal"
    WORK = "work"


class Email(ExtendedDocument, ExpiredAtDoc):
    address: EmailStr = Field(..., description="The email address")
    type: Optional[EmailType] = Field(default=None, description="The email type")
    current: Optional[bool] = Field(
        default=None, description="Whether the email is the current one"
    )
    pdl_code: Optional[str] = Field(default=None, description="The pdl code")
    provenance: Optional[Provenance] = Field(default=None, description="The provenance")

    class Settings:
        equality_fields = ["address"]

    @property
    def domain(self) -> str:
        return self.address.split("@")[1]

    @property
    def name(self) -> str:
        return self.address.split("@")[0]


class UnlinkedEmail(CreatedAtDoc):
    """Used for emails that are not linked to a person."""

    address: EmailStr = Field(..., description="The email address")
    email_type: EmailType = Field(default=EmailType.OTHER, description="The email type")
    provenance: Optional[Provenance] = Field(default=None, description="The provenance")

    class Settings:
        collection_name = "unlinked_emails"
        equality_fields = ["address"]



from typing import TYPE_CHECKING, Any, Optional

from beanie import PydanticObjectId
from pydantic import Field

from qai.schema.models.addons import (
    CreatedAtDoc,
    DateRange,
    Deleteable,
    Labels,
    Taggable,
)
from qai.schema.models.address_model import Address
from qai.schema.models.company_model import Company
from qai.schema.models.email_model import Email
from qai.schema.models.models import ExcludeReason
from qai.schema.models.name_model import Name
from qai.schema.models.person_model import Person
from qai.schema.models.phone_number_model import PhoneNumber
from qai.schema.models.qbeanie import Link
from qai.schema.models.social_media_model import SocialMedia

if TYPE_CHECKING:
    from qai.schema.models.campaign_model import Campaign


class Contact(CreatedAtDoc, Taggable, Labels, Deleteable):
    """A snapshot in time of a person for outreach purposes."""

    person: Link[Person] = Field(..., description="The reference to the person")
    person_id: PydanticObjectId = Field(..., description="The ID of the person")
    name: Name = Field(..., description="The name of the contact")
    location: Optional[Address] = Field(default=None, description="The location of the contact")
    campaign_id: PydanticObjectId = Field(..., description="The ID of the campaign")
    company: Optional[Link[Company]] = Field(
        default=None, description="The reference to the company"
    )
    email: Optional[Email] = Field(default=None, description="The email of the contact")
    email_verifications: Optional[dict[str, dict[str, Any]]] = Field(
        default=None, description="The email verifications of the contact"
    )
    phone: Optional[PhoneNumber] = Field(
        default=None, description="The phone number of the contact"
    )
    social_media: Optional[list[SocialMedia]] = Field(
        default=None, description="The social media urls of the person"
    )
    notes: Optional[str] = Field(default=None, description="The notes of the contact")
    excluded: Optional[bool] = Field(default=None, description="Whether the contact is excluded")
    excluded_reasons: Optional[list[ExcludeReason | str]] = Field(
        default=None, description="The reason for the exclusion"
    )
    temp_pdl: Optional[dict[str, Any]] = Field(
        default=None, description="The temporary PDL data of the person"
    )
    temp_nubela: Optional[dict[str, Any]] = Field(
        default=None, description="The temporary Nubela data of the person"
    )

    temp_linkedin: Optional[dict[str, Any]] = Field(
        default=None, description="The temporary LinkedIn data of the person"
    )

    temp_linkedin_activity: Optional[dict[str, Any]] = Field(
        default=None, description="The temporary LinkedIn data of the person"
    )

    class Settings:
        name = "contacts"
        equality_fields = ["person_id", "campaign_id"]
        keep_nulls = False

    @staticmethod
    def from_data(person: Person, company: Company, campaign: "Campaign") -> "Contact":
        if not person.id:
            raise ValueError(f"Person ID is required. Person: {person}")
        if not campaign.id:
            raise ValueError(f"Campaign ID is required. Campaign: {campaign}")
        c = Contact(
            person=person,
            person_id=person.id,
            name=person.name if person.name else Name(),
            campaign_id=campaign.id,
            company=company,
            email=person.emails[0] if person.emails else None,
            phone=person.phone_numbers[0] if person.phone_numbers else None,
            social_media=person.social_media,
            location=person.addresses[0] if person.addresses else None,
        )
        return c

    @classmethod
    async def get_or_create(
        cls: type["Contact"], person: Person, company: Company, campaign: "Campaign"
    ) -> "Contact":
        contact = cls.from_data(person=person, company=company, campaign=campaign)
        return await Contact.find_or_insert(contact)

    @property
    def safe_company_name(self) -> str:
        return self.company.name if self.company else ""

class ContactList(CreatedAtDoc, Deleteable, DateRange, Taggable, Labels):
    name: str = Field(..., description="The name of the contact list")
    description: Optional[str] = Field(
        default=None, description="The description of the contact list"
    )

    contacts: list[Link[Contact]] = Field(
        default=None, description="The references to the people of the contact list"
    )

    class Settings:
        name = "contact_lists"
        equality_fields = ["name"]
        keep_nulls = False



import re
from enum import StrEnum


class MatchingAlgorithm(StrEnum):
    BAG_OF_WORDS = "Bag of Words"
    EXACT = "Exact"
    REGEX = "Regex"
    CONTAINS = "Contains"


def clean_domain(domain: str) -> str:
    # Remove protocol if present
    domain = re.sub(r"^https?://", "", domain)

    # Remove www. if present
    domain = re.sub(r"^www\.", "", domain)

    # Remove any trailing slash
    domain = domain.rstrip("/")

    return domain


def title_matches(
    title: str,
    filter_titles: list[str],
    matching_algorithms: list[MatchingAlgorithm] = [
        MatchingAlgorithm.BAG_OF_WORDS,
        MatchingAlgorithm.CONTAINS,
    ],
) -> bool:
    if not title:
        return False
    filter_titles_bow = [set(title.split(" ")) for title in filter_titles]

    for algorithm in matching_algorithms:
        if algorithm == MatchingAlgorithm.BAG_OF_WORDS:
            job_title = set(title.split(" "))
            if any(job_title.issubset(title) for title in filter_titles_bow):
                return True
        elif algorithm == MatchingAlgorithm.EXACT:
            if title in filter_titles:
                return True
        elif algorithm == MatchingAlgorithm.REGEX:
            if any(re.search(title, title, re.IGNORECASE) for title in filter_titles):
                return True
        elif algorithm == MatchingAlgorithm.CONTAINS:
            if any(title in title for title in filter_titles):
                return True
    return False



import re
from typing import Any, Optional

from beanie import PydanticObjectId
from pydantic import Field
from qai.schema.models.addons import (
    CreatedAtDoc,
    Deleteable,
    Labels,
    Provenance,
    Taggable,
)
from qai.schema.models.address_model import Address
from qai.schema.models.email_model import Email
from qai.schema.models.job_model import Job, _sort_job_key
from qai.schema.models.models import GenderEnum
from qai.schema.models.name_model import Name
from qai.schema.models.phone_number_model import PhoneNumber
from qai.schema.models.social_media_model import SocialMedia, SocialMediaType


class Person(CreatedAtDoc, Deleteable, Taggable, Labels):
    id: Optional[PydanticObjectId] = Field(
        default=None, description="The unique identifier of the person"
    )
    name: Optional[Name] = Field(default=None, description="The name of the person")
    gender: Optional[GenderEnum] = Field(
        default=None, description="The gender identity of the person"
    )
    emails: Optional[list[Email]] = Field(default=None, description="The emails of the person")
    social_media: Optional[list[SocialMedia]] = Field(
        default=None, description="The social media urls of the person"
    )

    phone_numbers: Optional[list[PhoneNumber]] = Field(
        default=None, description="The phone numbers of the person"
    )
    addresses: Optional[list[Address]] = Field(
        default=None, description="The addresses of the person"
    )

    work_history: Optional[list[Job]] = Field(
        default=None, description="The work history of the person"
    )
    education_history: Optional[list[Job]] = Field(
        default=None, description="The education history of the person"
    )
    sources: Optional[list[Provenance]] = Field(
        default=None, description="The sources of the person"
    )
    provenance: Optional[Provenance] = Field(
        default=None, description="The provenance of the person"
    )
    activities: Optional[list[dict[Any, Any]]] = Field(
        default=None, description="The online activities of the person"
    )
    skills: Optional[list[str]] = Field(default=None, description="The skills of the person")

    additional_data: Optional[dict[str, Any]] = Field(
        default=None, description="Additional data of the person, a catch-all field"
    )

    class Settings:
        name = "people"
        equality_fields = ["id"]
        keep_nulls = False

    @property
    def full_name(self) -> str:
        if self.name:
            return self.name.full_name
        return ""

    @property
    def work_email(self) -> Optional[str]:
        work_email = self.get_work_email()
        if work_email:
            return work_email.address
        if self.emails:
            for email in self.emails:
                if email.type == "work":
                    return email.address
        return None

    @property
    def work_title(self) -> Optional[str]:
        work_title = self.get_work_title()
        if work_title:
            return work_title
        return None

    @property
    def job(self) -> Job:
        if self.work_history:
            return self.work_history[0]
        raise ValueError("No job found for person")

    def get_linkedin_url(self) -> Optional[str]:
        if not self.social_media:
            return None
        for sm in self.social_media:
            if sm.type == SocialMediaType.LINKEDIN:
                return sm.url
        return None

    def get_work_email(self) -> Optional[Email]:
        if not self.work_history:
            return None

        work_history = sorted(self.work_history, key=_sort_job_key, reverse=True)
        return work_history[0].email

    def get_work_title(self) -> Optional[str]:
        if not self.work_history:
            return None
        ## TODO: slintel doesn't have a start or end date
        ## so we can't sort by start date or end date for now
        # work_history = sorted(self.work_history, key=_sort_job_key, reverse=True)
        # return work_history[0].title
        return self.work_history[0].title

    def match_query(self) -> dict[str, Any]:
        query: list[dict[str, Any]] = []

        # Match by email
        if self.emails and isinstance(self.emails, list):
            email_conditions = [
                {"emails.address": re.compile(f"^{re.escape(email.address)}$", re.IGNORECASE)}
                for email in self.emails
                if email.address
            ]
            if email_conditions:
                query.append({"$or": email_conditions})

        # Match by name and address
        if self.name and self.addresses and isinstance(self.addresses, list):
            for address in self.addresses:
                address_conditions = {}
                if self.name.first:
                    address_conditions["name.first"] = re.compile(
                        f"^{re.escape(self.name.first)}$", re.IGNORECASE
                    )
                if self.name.last:
                    address_conditions["name.last"] = re.compile(
                        f"^{re.escape(self.name.last)}$", re.IGNORECASE
                    )
                if address.street:
                    address_conditions["addresses.street"] = re.compile(
                        f"^{re.escape(address.street)}$", re.IGNORECASE
                    )
                if address.city:
                    address_conditions["addresses.city"] = re.compile(
                        f"^{re.escape(address.city)}$", re.IGNORECASE
                    )
                if address.state:
                    address_conditions["addresses.state"] = re.compile(
                        f"^{re.escape(address.state)}$", re.IGNORECASE
                    )
                if address.postal_code:
                    address_conditions["addresses.postal_code"] = address.postal_code
                if address.country:
                    address_conditions["addresses.country"] = re.compile(
                        f"^{re.escape(address.country)}$", re.IGNORECASE
                    )

                if len(address_conditions) > 0:
                    query.append({"$and": [{k: v} for k, v in address_conditions.items()]})
        ## TODO - clean up match conditions
        # # Match by name and job company ID
        # if self.name and self.work_history and isinstance(self.work_history, list):
        #     for job in self.work_history:
        #         if job.company_id and self.name.first and self.name.last:
        #             query.append({
        #                 "$and": [
        #                     {"name.first": re.compile(f"^{re.escape(self.name.first)}$", re.IGNORECASE)},
        #                     {"name.last": re.compile(f"^{re.escape(self.name.last)}$", re.IGNORECASE)},
        #                     {"work_history.company_id": job.company_id}
        #                 ]
        #             })

        # Match by social media profiles
        if self.social_media and isinstance(self.social_media, list):
            for sm in self.social_media:
                if sm.url and sm.type:
                    query.append(
                        {
                            "$and": [
                                {"social_media.url": sm.url},
                                {"social_media.type": str(sm.type)},
                            ]
                        }
                    )

        # # Match by phone number
        # if self.phone_numbers and isinstance(self.phone_numbers, list):
        #     for phone in self.phone_numbers:
        #         if phone.number and phone.type:
        #             query.append({
        #                 "$and": [
        #                     {"phone_numbers.number": phone.number},
        #                     {"phone_numbers.type": phone.type}
        #                 ]
        #             })

        if not query:
            raise ValueError(f"Couldn't create match_query for person {self}")

        q = {"$or": query}
        return q



from datetime import datetime
from typing import Optional

from beanie import PydanticObjectId
from flexible_datetime import flextime
from pydantic import BaseModel, Field

from qai.schema.extensions import ExtendedDocument
from qai.schema.mergers.merge import (
    HIGH_PRIORITY,
    HIGHER_PRIORITY,
    LOW_PRIORITY,
    LOWER_PRIORITY,
    NORMAL_PRIORITY,
)

priorities = {
    "manual": NORMAL_PRIORITY,
    "ai_generated": LOW_PRIORITY,
    "linkedin": HIGHER_PRIORITY,
    "slintel": HIGH_PRIORITY,
    "pdl": HIGH_PRIORITY,
    "nubela": HIGH_PRIORITY,
}


def get_priority(source: str):
    return priorities.get(source.lower(), NORMAL_PRIORITY)


class Taggable(BaseModel):
    tags: Optional[list[str]] = Field(default=None, description="The tags for the document")


class Labels(BaseModel):
    tags: Optional[list[str]] = Field(default=None, description="The tags for the document")


class ExpiredAtDoc(BaseModel):
    expired_at: Optional[flextime] = Field(
        default=None, description="The timestamp of the document expiration"
    )


class Provenance(BaseModel):
    source: str = Field(default=None, description="The source of the document")
    source_id: Optional[PydanticObjectId] = Field(
        default=None, description="The source id of the document"
    )
    str_id: Optional[str] = Field(
        default=None, description="The source id of the document if not inserted into the database"
    )
    sources: Optional[list[str]] = Field(
        default=None, description="The sources of the document"
    )

    @property
    def priority(self):
        return get_priority(self.source)


class Deleteable(BaseModel):
    deleted_at: Optional[datetime] = Field(
        default=None, description="The timestamp of the document deletion"
    )
    is_deleted: bool = Field(default=False, description="Whether the document is deleted")


class DateRange(BaseModel):
    start: Optional[flextime] = Field(default=None, description="The start date of the range")
    end: Optional[flextime] = Field(default=None, description="The end date of the range")


class Updateable(BaseModel):
    updated_at: Optional[flextime] = Field(
        default=None, description="The timestamp of the document update"
    )
    last_checked_at: Optional[flextime] = Field(
        default=None, description="The timestamp of the last check"
    )


class CreatedAtDoc(ExtendedDocument):
    created_at: flextime = Field(
        default_factory=flextime, description="The timestamp of the document creation"
    )



from enum import StrEnum
from typing import TYPE_CHECKING, Any, Optional, Self, cast

from pydantic import Field, field_validator
from qai.schema.models.addons import CreatedAtDoc, Deleteable, Labels, Taggable
from qai.schema.models.address_model import Address
from qai.schema.models.email_model import Email
from qai.schema.models.person_model import Person
from qai.schema.models.phone_number_model import PhoneNumber
from qai.schema.models.qbeanie import Link
from qai.schema.models.social_media_model import SocialMedia, SocialMediaType
from qai.schema.utils.utils import clean_domain


class CompanyType(StrEnum):
    PUBLIC = "public"
    PRIVATE = "private"
    EDUCATIONAL = "educational"
    NON_PROFIT = "non_profit"
    GOVERNMENT = "government"
    SELF_EMPLOYED = "self_employed"
    PARTNERSHIP = "partnership"
    SOLE_PROPRIETORSHIP = "sole_proprietorship"
    OTHER = "other"


class Company(CreatedAtDoc, Taggable, Labels, Deleteable):
    name: str = Field(..., description="The name of the company")
    domains: list[str] = Field(default_factory=list, description="The websites of the company")
    business_name: Optional[str] = Field(
        default=None,
        description="The official business name of the company, government registered.",
    )
    company_type: Optional[CompanyType] = Field(
        default=None, description="The type of the company"
    )
    description: Optional[str] = Field(default=None, description="The description of the company")
    tagline: Optional[str] = Field(default=None, description="The tagline of the company")
    parent_company: Optional[Self] = None
    linkedin_id: Optional[int] = Field(default=None, description="The LinkedIn ID of the company")
    people: list[Link[Person]] = Field(
        default=None, description="The references to the people of the company"
    )
    emails: Optional[list[Email]] = Field(default=None, description="The emails of the company")
    is_branch: Optional[bool] = Field(default=None, description="Whether the company is a branch")
    is_entity: Optional[bool] = Field(default=None, description="Whether the company is an entity")
    location: Optional[Address] = Field(
        default=None, description="The location of the company's headquarters"
    )
    industries: Optional[list[str]] = Field(
        default=None, description="The industries of the company"
    )
    specialties: Optional[list[str]] = Field(
        default=None, description="The specialties of the company"
    )
    employee_range: Optional[str] = Field(
        default=None, description="The size range of employees in the company"
    )

    addresses: Optional[list[Address]] = Field(
        default=None, description="The addresses of the company"
    )
    social_media: Optional[list[SocialMedia]] = Field(
        default=None, description="The social media urls of the person"
    )
    phone_numbers: Optional[list[PhoneNumber]] = Field(
        default=None, description="The phone numbers of the company"
    )
    revenues: Optional[dict[str, int]] = Field(
        default=None, description="The revenues of the company per year"
    )
    ticker: Optional[str] = Field(default=None, description="The company ticker")
    founded: Optional[int] = Field(default=None, description="The founding year of the company")
    existing_customers: Optional[list["Company"]] = Field(
        default=None, description="The existing customers of the company"
    )

    class Settings:
        name = "companies"
        equality_fields = ["id"]
        keep_nulls = False

    def match_query(self) -> dict[str, Any]:
        query: list[dict[str, Any]] = []

        # Match by ID
        if self.id:
            query.append({"_id": self.id})

        # Match by name and domains
        if self.name and self.domains:
            query.append({"$and": [{"name": self.name}, {"domains": {"$in": self.domains}}]})

        # Match by LinkedIn URL
        if self.social_media:
            linkedin_urls = [
                sm.url for sm in self.social_media if sm.type == SocialMediaType.LINKEDIN and sm.url
            ]
            if linkedin_urls:
                query.append({"$or": [{"social_media.url": url} for url in linkedin_urls]})

        # Match by LinkedIn ID
        if self.linkedin_id:
            query.append({"linkedin_id": self.linkedin_id})

        if not query:
            raise ValueError(f"Couldn't create match_query for company {self}")

        return {"$or": query}

    @property
    def website(self) -> str:
        if self.domains:
            return self.domains[0]
        raise ValueError(f"No website for company domains={self.domains}. ")

    @classmethod
    def find_by_linkedin_url(cls, linkedin_url: str) -> dict[str, Any]:
        return {"social_media.url": linkedin_url}

    @field_validator("domains", mode="before")
    def strip_protocols(cls, v):
        if isinstance(v, list):
            return [clean_domain(domain) for domain in v]
        return v

    def eq(self, other, nones_ok: bool = False, to_lower: bool = False):
        """Equality check for Company.
        Business names can be different, so we only check the name and domains.
        domains are checked first because they are more likely to be unique.
        after that, we check the name.
        """
        if not super().eq(other):
            return False
        other = cast(Company, other)
        if not self.domains and not other.domains:
            return self.name == other.name and self.domains == other.domains
        return True

    def full_print(self):
        """Print the company with all its fields"""
        return self.model_dump(by_alias=True, exclude_none=True)



from typing import Generic, TypeVar

from pydantic import BaseModel

from .merge import Priority, merge_value

T = TypeVar("T")

# SourcePriority is now just an alias for int


def merge_companies(
    target: BaseModel, source: BaseModel, target_priority: Priority, source_priority: Priority
) -> None:
    for field_name, field in target.model_fields.items():
        if hasattr(source, field_name):
            target_value = getattr(target, field_name)
            source_value = getattr(source, field_name)

            if target_value is not None or source_value is not None:
                merged_value = merge_value(
                    target_value, source_value, target_priority, source_priority
                )
                setattr(target, field_name, merged_value)



import importlib
import re
from logging import getLogger
from typing import Any, Callable, Optional, TypeVar, cast

from addressformatting import AddressFormatter  # type: ignore
from pydantic import Field, field_validator
from qai.schema.extensions import ExtendedDocument
from qai.schema.models.addons import Provenance

log = getLogger(__name__)

T = TypeVar("T", bound="Address")


class Address(ExtendedDocument):
    street: Optional[str] = Field(default=None, description="The street address")
    street2: Optional[str] = Field(default=None, description="The street address")
    city: Optional[str] = Field(default=None, description="The city")
    state: Optional[str] = Field(default=None, description="The state")
    postal_code: Optional[str] = Field(default=None, description="The postal code")
    country: Optional[str] = Field(default=None, description="The country")
    current: Optional[bool] = Field(
        default=None, description="Whether the address is the current one"
    )
    notes: Optional[str] = Field(default=None, description="The address notes")
    provenance: Optional[Provenance] = Field(
        default=None, description="The provenance of the address"
    )
    raw: Optional[str] = Field(default=None, description="The raw address, to be parsed")

    def equality_hash(self):
        ## make a tuple of all equality_fields
        all_kv = self.model_fields.items()
        return tuple([getattr(self, k) for k, v in all_kv if k in self.Settings.equality_fields])

    @field_validator("street", "street2", "city", "state", "postal_code", "country", "notes", "raw")
    @classmethod
    def strip_whitespace(cls, v: Optional[str]) -> Optional[str]:
        return v.strip() if v else v

    @classmethod
    def from_str(cls: type[T], s: str, *args, **kwargs) -> T:
        return cls.parse_address(s=s, *args, **kwargs)  # type: ignore

    @classmethod
    def from_dict_mapping(cls, data: dict[str, Any], mapping: dict[str, str]) -> "Address":
        d = {
            address_field: data[dict_key]
            for dict_key, address_field in mapping.items()
            if dict_key in data
        }
        return cls(**d)

    @classmethod
    def parse_address(cls: type["T"], s: str, *args, **kwargs) -> T:
        if not hasattr(cls, "_parse_address"):
            setattr(cls, "_parse_address", _load_address_parsers())
            assert cls._parse_address, "No address parser loaded"  # type: ignore
        try:
            return cls._parse_address(s=s, *args, **kwargs)  # type: ignore
        except:
            return cast(T, Address(raw=s))

    class Settings:
        equality_fields = ["street", "street2", "city", "state", "postal_code", "country"]

    def __str__(self) -> str:
        return self.format_address(self)

    @classmethod
    def format_address(cls, address: "Address", one_line: bool = False) -> str:
        address_dict = {
            "road": address.street,
            "house": address.street2,
            "city": address.city,
            "state": address.state,
            "postcode": address.postal_code,
            "country": address.country,
        }

        # Remove None values
        data_dict = {k: v for k, v in address_dict.items() if v is not None}
        formatter = AddressFormatter()
        # Format the address using address-formatting
        if one_line:
            formatted_address = formatter.one_line(data_dict, country=address.country)  # type: ignore
        else:
            formatted_address = formatter.format(data_dict, country=address.country)  # type: ignore
        # Ensure state is included if it's in the original address
        if address.state and address.state not in formatted_address:
            if one_line:
                parts = formatted_address.split(", ")
                if len(parts) >= 2:
                    parts.insert(-1, address.state)
                formatted_address = ", ".join(parts)
            else:
                lines = formatted_address.split("\n")
                if len(lines) >= 2:
                    lines.insert(-1, address.state)
                formatted_address = "\n".join(lines)

        # Remove any whitespace before or after newlines, and strip the entire string
        formatted_address = re.sub(r"\s*\n\s*", "\n", formatted_address.strip())
        if not formatted_address:
            formatted_address = address.raw or ""
        return formatted_address.strip()


def _load_address_parser(module_path: str) -> Optional[Callable]:
    try:
        module = importlib.import_module(module_path)
        parse_address = getattr(module, "parse_address")
        return parse_address
    except ImportError:
        log.debug(f"Failed to import {module_path} parser")
    except AttributeError:
        log.debug(f"{module_path} parser does not have parse_address function")
    return None


def _basic_parse_address(s: str, *args, **kwargs) -> Address:
    return Address(raw=s)


def _load_address_parsers() -> Callable:
    log.debug("Loading external Address parsers")

    parsers = [
        ("postal", "qai.schema.parsers.address_parser_postal"),
        ("pyap", "qai.schema.parsers.address_parser_pyap"),
    ]

    for parser_name, module_path in parsers:
        parse_address = _load_address_parser(module_path)
        if parse_address:
            log.debug(f"Loaded {parser_name} parser")
            return parse_address
    else:
        log.error("No Address parser could be loaded")
        return _basic_parse_address



from enum import StrEnum
from typing import Optional

from pydantic import BaseModel, Field

from qai.schema.models.addons import Provenance


class SocialMediaType(StrEnum):
    OTHER = "other"
    LINKEDIN = "linkedin"
    TWITTER = "twitter"
    FACEBOOK = "facebook"
    INSTAGRAM = "instagram"
    GITHUB = "github"


class SocialMedia(BaseModel):
    url: str = Field(..., description="The social media url")
    type: SocialMediaType = Field(
        default=SocialMediaType.OTHER, description="The social media type"
    )
    provenance: Optional[Provenance] = Field(
        default=None, description="The provenance of the address"
    )

    class Settings:
        equality_fields = ["url"]



from qai.schema.extensions import DocExtensions, ExtendedDocument
from qai.schema.models.addons import Provenance
from qai.schema.models.address_model import Address
from qai.schema.models.campaign_model import (
    Campaign,
    CampaignBatch,
    CampaignOptions,
    CampaignStep,
)
from qai.schema.models.company_model import Company, CompanyType
from qai.schema.models.contact_model import Contact, ContactList
from qai.schema.models.email_model import Email, EmailType
from qai.schema.models.existing_customer_model import ExistingCustomer
from qai.schema.models.job_model import Job
from qai.schema.models.models import ExcludeReason, GenderEnum, OutreachType
from qai.schema.models.name_model import Name
from qai.schema.models.person_model import Person
from qai.schema.models.phone_number_model import PhoneNumber, PhoneType
from qai.schema.models.social_media_model import SocialMedia, SocialMediaType

__all__ = [
    "DocExtensions",
    "ExtendedDocument",
    "Provenance",
    "Address",
    "Campaign",
    "CampaignBatch",
    "CampaignOptions",
    "CampaignStep",
    "Company",
    "CompanyType",
    "Contact",
    "ContactList",
    "Email",
    "EmailType",
    "ExistingCustomer",
    "Job",
    "ExcludeReason",
    "GenderEnum",
    "OutreachType",
    "Name",
    "Person",
    "PhoneNumber",
    "PhoneType",
    "SocialMedia",
    "SocialMediaType",
]



from enum import StrEnum
from typing import Optional

from pydantic import Field

from qai.schema.models.addons import CreatedAtDoc, Provenance


class PhoneType(StrEnum):
    OTHER = "other"
    MOBILE = "mobile"
    HOME = "home"
    WORK = "work"
    TOLL_FREE = "toll_free"
    FAX = "fax"


class PhoneNumber(CreatedAtDoc):
    number: str = Field(..., description="The phone number")
    type: PhoneType = Field(default=PhoneType.OTHER, description="The phone type")
    notes: Optional[str] = Field(default=None, description="The phone number notes")
    provenance: Optional[Provenance] = Field(
        default=None, description="The provenance of the PhoneNumber"
    )

    class Settings:
        equality_fields = ["number"]



from typing import TYPE_CHECKING

from beanie import PydanticObjectId
from pydantic import Field

from qai.schema.extensions import ExtendedDocument

from qai.schema.models.addons import Deleteable, Labels, Taggable
from qai.schema.models.qbeanie import Link

if TYPE_CHECKING:
    from .company_model import Company


class ExistingCustomer(ExtendedDocument, Taggable, Labels, Deleteable):
    company: Link["Company"] = Field(..., description="The reference to the company")
    company_id: PydanticObjectId = Field(..., description="The ID of the company")
    name: str = Field(..., description="The name of the company")

    class Settings:
        name = "existing-customers"
        equality_fields = ["company_id"]
        keep_nulls = False



""" Pyap doesn't support addresses from all countries, 
and seems limited in its parsing capabilities. But it's still better than nothing.
"""

import re
from typing import Any, TypeVar, cast

import pyap

from qai.schema.models.address_model import Address

ET = TypeVar("ET", bound="Address")
country_map = {"United States": "US", "Great Britain": "GB", "United Kingdom": "GB", "Canada": "CA"}


def parse_address(s: str, country: str = "US", *args, **kwargs) -> Address:
    # Regex pattern to match country at the end of the string
    original_s = s
    pattern = r",\s*({})\s*$".format("|".join(country_map.keys()))

    match = re.search(pattern, s, re.IGNORECASE)

    if match:
        country = match.group(1)
        country = country_map[country.title()]
        s = re.sub(pattern, "", s, flags=re.IGNORECASE).strip()

    # Parse the address using pyap
    fake_street = ""
    parsed_addresses = pyap.parse(s, country=country)

    if not parsed_addresses:
        ## Try adding a fake street number to the address
        fake_street = "123 Fake St, "
        s = f"{fake_street}{s}"
        parsed_addresses = pyap.parse(s, country=country)

    if not parsed_addresses:
        raise ValueError(f"Unable to parse the address {s}, country={country}")

    # Take the first parsed address (assuming there's only one)
    parsed_address: Any = parsed_addresses[0]

    # Get the parsed address as a dictionary
    address_dict = cast(dict[str, Any], parsed_address.data_as_dict)

    if fake_street:
        # Remove the fake street number
        address_dict.pop("full_street", None)
        address_dict.pop("street_name", None)
        address_dict.pop("street_type", None)
        address_dict.pop("street_number", None)
        address_dict.pop("post_direction", None)
        address_dict["full_address"] = address_dict.get("full_address", "").replace(fake_street, "", 1)
        

    # Map the parsed fields to our Address class
    address = Address(
        street=address_dict.get("full_street"),
        city=address_dict.get("city"),
        state=address_dict.get("region1"),  # pyap uses 'region1' for state
        postal_code=address_dict.get("postal_code"),
        country=address_dict.get("country_id"),
        raw=original_s,
    )
    return address



from typing import TYPE_CHECKING, Optional, TypeAlias, TypeVar

from beanie import Document
from beanie import Link as BeanieLink
from beanie import PydanticObjectId
from beanie.odm.queries.find import FindOne as BeanieFindOne
from pydantic import BaseModel, EmailStr, Field, field_validator
from qai.schema.models.addons import (
    CreatedAtDoc,
    DateRange,
    Deleteable,
    Labels,
    Taggable,
)
from qai.schema.models.company_model import Company
from qai.schema.models.contact_model import Contact
from qai.schema.models.models import OutreachType
from qai.schema.models.qbeanie import Link


class EnrichmentSources(BaseModel):
    pdl: bool = False
    nubela: bool = False
    linkedin: bool = False


class CampaignOptions(BaseModel):
    skip_existing_customers: bool = Field(default=True, title="Skip existing customers")


class Campaign(CreatedAtDoc, Deleteable, DateRange, Taggable, Labels):
    name: str = Field(..., description="The name of the campaign")
    description: Optional[str] = Field(default=None, description="The description of the campaign")

    contacts: list[Link[Contact]] = Field(
        default=None, description="The references to the people of the campaign"
    )
    companies: list[Link[Company]] = Field(
        default=None, description="The references to the companies of the campaign"
    )
    campaign_steps: list[Link["CampaignStep"]] = Field(default_factory=list)

    exclude_domains: list[str] = Field(default_factory=list, title="List of domains to filter out")
    exclude_emails: list[EmailStr] = Field(
        default_factory=list, title="List of emails to filter out"
    )
    exclude_titles: list[str] = Field(default_factory=list, title="List of titles to filter out")
    include_emails: list[str] = Field(default_factory=list, title="List of emails to include")
    include_titles: list[str] = Field(default_factory=list, title="List of titles to include")

    ## TODO information about Ideal Customer Profile (ICP)

    class Settings:
        name = "campaigns"
        equality_fields = ["name"]
        keep_nulls = False

    @field_validator("exclude_domains", mode="before")
    def strip_protocols(cls, v):
        if isinstance(v, list):
            v = [domain.lstrip("http://").lstrip("https://").lstrip("www.") for domain in v]
        return v


class CampaignStep(CreatedAtDoc, Deleteable, DateRange, Taggable, Labels):
    campaign_id: PydanticObjectId = Field(..., description="The ID of the campaign")
    name: str = Field(..., description="The name of the step")
    step: int = Field(..., description="The order of the step")
    description: Optional[str] = Field(default=None, description="The description of the step")
    outreach_type: Optional[OutreachType] = Field(default=None, description="The type of outreach")
    contacts: list[Link[Contact]] = Field(
        default=None, description="The references to the people of the step"
    )
    # companies: list[Link[Company]] = Field(
    #     default=None, description="The references to the companies of the step"
    # )
    batches: list[Link["CampaignBatch"]] = Field(
        default_factory=list, description="The references to the batches of the step"
    )
    notes: Optional[str] = Field(default=None, description="The notes of the batch")

    class Settings:
        name = "campaign_steps"
        equality_fields = ["campaign_id", "name"]
        keep_nulls = False


class CampaignBatch(CreatedAtDoc, Deleteable, DateRange, Taggable, Labels):
    campaign_id: PydanticObjectId = Field(..., description="The ID of the campaign")
    campaign_step_id: PydanticObjectId = Field(..., description="The ID of the campaign step")
    order: int = Field(..., description="The order of the batch")
    name: str = Field(..., description="The name of the batch")
    contacts: list[Link[Contact]] = Field(
        default=None, description="The references to the people of the batch"
    )
    notes: Optional[str] = Field(default=None, description="The notes of the batch")

    class Settings:
        name = "campaign_batches"
        equality_fields = ["campaign_id", "campaign_step_id", "order", "name"]
        keep_nulls = False



import asyncio
import json
from abc import abstractmethod
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Self,
    Sequence,
    Type,
    TypeVar,
    Union,
)
from urllib.parse import urlparse

from beanie import Document, View
from beanie.exceptions import CollectionWasNotInitialized
from beanie.odm.actions import ActionDirections, EventTypes, wrap_with_actions
from beanie.odm.documents import DocType, DocumentProjectionType
from beanie.odm.fields import WriteRules
from beanie.odm.settings.document import DocumentSettings
from beanie.odm.utils.self_validation import validate_self_before
from beanie.odm.utils.state import save_state_after
from motor.motor_asyncio import AsyncIOMotorCollection
from pydantic._internal._model_construction import ModelMetaclass
from pymongo.client_session import ClientSession
from qai.schema.mergers.merge import NORMAL_PRIORITY, Priority, merge_model

T = TypeVar("T", bound=Document)
ET = TypeVar("ET", bound="ExtendedDocument")

offline_settings = DocumentSettings()

# Create a global registry dictionary
extended_document_registry: dict[str, Any] = {}


# Define the metaclass for registering subclasses
class ExtendedDocumentMeta(type):
    def __new__(cls, name, bases, dct):
        new_class = super().__new__(cls, name, bases, dct)
        extended_document_registry[name] = new_class
        return new_class


# Define a combined metaclass
class CombinedMeta(ExtendedDocumentMeta, ModelMetaclass):
    pass


class ExtendedDocument(Document, metaclass=CombinedMeta):

    @classmethod
    def get_motor_collection(cls) -> AsyncIOMotorCollection:
        try:
            return super().get_motor_collection()
        except CollectionWasNotInitialized:
            """We don't mind if the collection is not initialized
            until we actually need the collection for db interactions
            """
            return None  # type: ignore

    @classmethod
    async def load(
        cls: type[ET],
        projection_model: Type["DocumentProjectionType"] = None,  # type: ignore
        session: Optional[ClientSession] = None,
        ignore_cache: bool = False,
        fetch_links: bool = False,
        with_children: bool = False,
        nesting_depth: Optional[int] = None,
        nesting_depths_per_field: Optional[Dict[str, int]] = None,
        *args: Union[Mapping[str, Any], bool],
    ) -> Optional[ET]:
        raise NotImplementedError("load must be implemented in the subclass of Extended")
        if not args:
            fields = cls.Settings.equality_fields
            args = tuple([params])  # type: ignore
        doc = await cls.find_one(
            args[0],
            projection_model=projection_model,
            session=session,
            ignore_cache=ignore_cache,
            fetch_links=fetch_links,
            with_children=with_children,
            nesting_depth=nesting_depth,
            nesting_depths_per_field=nesting_depths_per_field,
        )
        return doc

    @classmethod
    async def find_many_fetch(cls, *args) -> list[Self]:
        if args:
            ecs = await cls.find_many(*args).to_list()
        else:
            ecs = await cls.find_all().to_list()

        await asyncio.gather(*[ec.fetch_all_links() for ec in ecs])
        return ecs

    @wrap_with_actions(EventTypes.SAVE)
    @save_state_after
    @validate_self_before
    async def upsert(  # type: ignore
        self: ET,
        *args: Union[Mapping[str, Any], bool],
        projection_model: Type["DocumentProjectionType"] = None,  # type: ignore
        ignore_cache: bool = False,
        fetch_links: bool = False,
        with_children: bool = False,
        nesting_depth: Optional[int] = None,
        nesting_depths_per_field: Optional[Dict[str, int]] = None,
        session: Optional[ClientSession] = None,
        link_rule: WriteRules = WriteRules.DO_NOTHING,
        ignore_revision: bool = False,
        save_kwargs: Optional[dict] = None,  # type: ignore
        pymongo_kwargs: Optional[dict] = None,
        skip_actions: Optional[List[Union[ActionDirections, str]]] = None,
    ) -> ET:
        """
        Upsert document with provided args
        """
        if pymongo_kwargs is None:
            pymongo_kwargs = {}
        if save_kwargs is None:
            save_kwargs = {}

        ## Get all Settings.equality_fields
        if not args:
            params = self.match_query()
            args = tuple([params])  # type: ignore
        if skip_actions:
            raise NotImplementedError("skip_actions is not implemented")
        doc = await self.find_one(
            args[0],
            projection_model=projection_model,
            session=session,
            ignore_cache=ignore_cache,
            fetch_links=fetch_links,
            with_children=with_children,
            nesting_depth=nesting_depth,
            nesting_depths_per_field=nesting_depths_per_field,
        )

        if doc:
            return doc  # type: ignore
        return await self.save(
            session=session,
            link_rule=link_rule,
            ignore_revision=ignore_revision,
            **pymongo_kwargs,
        )

    @classmethod
    async def find_or_insert(  # type: ignore
        cls: type[ET],
        document: ET,
        *args: Union[Mapping[str, Any], bool],
        projection_model: Type["DocumentProjectionType"] = None,  # type: ignore
        ignore_cache: bool = False,
        fetch_links: bool = False,
        with_children: bool = False,
        nesting_depth: Optional[int] = None,
        nesting_depths_per_field: Optional[Dict[str, int]] = None,
        session: Optional[ClientSession] = None,
        link_rule: WriteRules = WriteRules.DO_NOTHING,
        ignore_revision: bool = False,
        insert_kwargs: Optional[dict] = None,  # type: ignore
        find_kwargs: Optional[dict] = None,
        skip_actions: Optional[List[Union[ActionDirections, str]]] = None,
    ) -> ET:
        """
        find or insert document with provided args
        """
        if find_kwargs is None:
            find_kwargs = {}
        if insert_kwargs is None:
            insert_kwargs = {}

        ## Get all Settings.equality_fields
        if not args:
            params = document.match_query()
            args = tuple([params])  # type: ignore
        doc = await document.find_one(
            args[0],
            projection_model=projection_model,
            session=session,
            ignore_cache=ignore_cache,
            fetch_links=fetch_links,
            with_children=with_children,
            nesting_depth=nesting_depth,
            nesting_depths_per_field=nesting_depths_per_field,
            **find_kwargs
        )

        if doc:
            return doc
        return await document.insert(
            session=session,
            link_rule=link_rule,
            skip_actions=skip_actions,
            **insert_kwargs,
        )

    @classmethod
    def from_str(cls: type[ET], s: str, *args, **kwargs) -> ET:
        raise NotImplementedError("from_str must be implemented in the subclass of Extended")

    def __merge__(self: ET, other: ET, self_priority: Priority, other_priority: Priority) -> ET:
        """
        Merge self with other
        """
        merge_model(
            target=self, source=other, target_priority=self_priority, source_priority=other_priority
        )
        return self

    def merge(
        self: ET,
        other: ET,
        self_priority: Priority = NORMAL_PRIORITY,
        other_priority: Priority = NORMAL_PRIORITY,
    ) -> ET:
        """
        Merge self with other
        """
        return self.__merge__(other, self_priority, other_priority)

    def __str__(self) -> str:
        """
        Return a string representation of the document. doesn't return the None fields
        """
        fields = []
        for field in self.model_fields.items():
            if field[1] is not None:
                fields.append(f"{field[0]}={field[1]}")
        return ", ".join(fields)

    def eq(self, other, nones_ok: bool = False, to_lower: bool = False):
        """
        Check if two documents are equal.
        Rules:
            1) If both have an id, compare the id.
            2) If any of the equality fields are None, return False unless nones_ok
            3) Compare the fields in equality_fields.
        """
        if not isinstance(other, self.__class__):
            return False
        if self.id is not None and other.id is not None:
            return self.id == other.id
        if not hasattr(self, "Settings") or not hasattr(self.Settings, "equality_fields"):
            raise ValueError("To use eq() equality_fields must be defined in Settings")

        ## iterate through fileds in equality_fields and compare them
        for field_name in self.Settings.equality_fields:
            val = getattr(self, field_name)
            if val is None and not nones_ok:
                return False
            try:
                oval = getattr(other, field_name)
            except AttributeError:
                if not nones_ok:
                    return False
                oval = None
            if to_lower:
                if isinstance(val, str):
                    val = val.lower()
                if isinstance(oval, str):
                    oval = oval.lower()

            if val != oval:
                return False
        return True

    def eq_none(self):
        if not hasattr(self, "Settings") or not hasattr(self.Settings, "equality_fields"):
            raise ValueError("To use eq_none() equality_fields must be defined in Settings")
        return all(
            getattr(self, field_name) is None for field_name in self.Settings.equality_fields
        )

    def match_query(self) -> dict[str, Any]:
        """
        Return a query that will match this document.
        """
        if not hasattr(self, "Settings") or not hasattr(self.Settings, "equality_fields"):
            raise ValueError("To use match_query() equality_fields must be defined in Settings")
        query = {}
        for field_name in self.Settings.equality_fields:
            query[field_name] = getattr(self, field_name)
        return query

    @staticmethod
    def get_base_url(url: str) -> str:
        if not url.startswith(("http://", "https://")):
            url = "http://" + url
        parsed_url = urlparse(url)

        # Get the hostname from parsed_url
        hostname = parsed_url.hostname
        if not hostname:
            raise ValueError(f"get_base_url: can't parse Invalid URL {url}")
        base_url = hostname[4:] if hostname.startswith("www.") else hostname

        return base_url

    async def load_self(self: ET) -> Optional[ET]:
        """
        ** Warning: This will overwrite any changes made to the document **
        Load self from the database
        """
        params = self.match_query()
        if not params:
            return None
        return await self.find_one(**params)

    def pformat(
        self: ET,
        indent: int = 2,
        exclude_none: bool = True,
        **kwargs,
    ) -> str:
        """
        Pretty format the document
        """
        return json.dumps(self.model_dump(exclude_none=True, **kwargs), indent=indent)


setattr(Document, "upsert", ExtendedDocument.upsert)


class DocExtensions:
    @wrap_with_actions(EventTypes.SAVE)
    @save_state_after
    @validate_self_before
    @abstractmethod
    async def upsert(  # type: ignore
        self: DocType,  # type: ignore
        *args: Union[Mapping[str, Any], bool],
        projection_model: Type["DocumentProjectionType"] = None,  # type: ignore
        ignore_cache: bool = False,
        fetch_links: bool = False,
        with_children: bool = False,
        nesting_depth: Optional[int] = None,
        nesting_depths_per_field: Optional[Dict[str, int]] = None,
        session: Optional[ClientSession] = None,
        link_rule: WriteRules = WriteRules.DO_NOTHING,
        ignore_revision: bool = False,
        save_kwargs: Optional[dict] = None,  # type: ignore
        pymongo_kwargs: Optional[dict] = None,
        skip_actions: Optional[list[ActionDirections | str]] = None,
    ) -> DocType: ...

    @abstractmethod
    def eq(self, other): ...

    @abstractmethod
    def eq_none(self): ...

    @staticmethod
    @abstractmethod
    def get_base_url(url): ...

    @classmethod
    @abstractmethod
    def from_str(cls, s: str): ...

    @abstractmethod
    def pformat(
        self: ET,
        indent: int = 2,
        exclude_none: bool = True,
        **kwargs,
    ) -> str: ...


def get_extended_documents(
    cls: type["Document"] = ExtendedDocument,
) -> list[type[Document] | type[View] | str]:
    return list(extended_document_registry.values())
